% This function runs the Substation Level State Estimation, and is called
% by central.m.
function [zcentral, H] = local(zlocal, localbranches, extlinedata, nodetype, s, display)

n_nodes = max(max(localbranches));

%Number of state variables:
n_V = n_nodes;
n_P = size(localbranches,1);
n_Q = n_P;

%Constructing measurement vectors:
z_V = zlocal(zlocal(:,3)==1 | zlocal(:,3)==11, :);
z_P= zlocal(zlocal(:,3)==2 | zlocal(:,3) == 4 | zlocal(:,3) == 6 | zlocal(:,3) == 12, :);
z_Q = zlocal(zlocal(:,3)==3 | zlocal(:,3) == 5 | zlocal(:,3) == 7 | zlocal(:,3) == 13, :);
nz_V = size(z_V,1);

%% Substation Level State Estimation, including BD processing
contP = true;
contQ = true;
contV = true;
while contP || contQ || contV
    
    if contP
        contP = false;
        nz_P = size(z_P,1);
        H.P =  zeros(nz_P, n_P);
        
        for k = 1:nz_P
            if z_P(k,3) == 2 || (z_P(k,3) == 4 && z_P(k,2) == 0) % if measurement type 2 and 4 (with external line)
                H.P(k, z_P(k,1) == localbranches(:,1))   = 1;
                H.P(k, z_P(k,1) == localbranches(:,2))   = -1;
            else % if measurement type 4 or 12
                H.P(k, (z_P(k,1) == localbranches(:,1)) & (z_P(k,2) == localbranches(:,2)))   = 1;
                H.P(k, (z_P(k,1) == localbranches(:,2)) & (z_P(k,2) == localbranches(:,1)))   = -1;
            end
        end
        
        %Orthogonal factorisation solution method
        zvec_P = z_P(:,4);
        Ginv2 = ( H.P' * ((diag(z_P(:,5))) \ H.P)) \ eye(n_P);
        Winvsq_P = diag(sqrt(z_P(:,5)));
        zvec_wP= Winvsq_P\zvec_P;
        [Q, R] = qr(Winvsq_P\H.P);
        Un = R(1:size(R,2),1:size(R,2));
        Qn = Q(:,1:size(R,2));
        zq = Qn'*zvec_wP;
        x2 = Un\zq;
        
        %Calculation of normalised residuals
        r_P = -zvec_P+H.P*x2;
        Ohm = diag(z_P(:,5))-H.P*Ginv2*H.P';
        r_P_N = abs(r_P)./sqrt(abs(diag(Ohm)));
        [r_P_N_max, i_P] = max(r_P_N);
        
    end

    if contQ 
        contQ = false;
        nz_Q = size(z_Q,1);
        H.Q =  zeros(nz_Q, n_Q);
        
        for k = 1:nz_Q
            if z_Q(k,3) == 3 || (z_Q(k,3) == 5 && z_Q(k,2) == 0) % if measurement type 3 and 5 (with external line)
                H.Q(k, z_Q(k,1) == localbranches(:,1))   = 1;
                H.Q(k, z_Q(k,1) == localbranches(:,2))   = -1;
            else % if measurement type 5 or type 13
                H.Q(k, (z_Q(k,1) == localbranches(:,1)) & (z_Q(k,2) == localbranches(:,2)))   = 1;
                H.Q(k, (z_Q(k,1) == localbranches(:,2)) & (z_Q(k,2) == localbranches(:,1)))   = -1;
            end
        end
        
        %Orthogonal factorisation solution method
        Ginv3 = ( H.Q' * ((diag(z_Q(:,5))) \ H.Q )) \ eye(n_Q);
        zvec_Q = z_Q(:,4);
        Winvsq_Q = diag(sqrt(z_Q(:,5)));
        zvec_wQ = Winvsq_Q\zvec_Q;
        [Q, R] = qr(Winvsq_Q\H.Q);
        Un = R(1:size(R,2),1:size(R,2));
        Qn = Q(:,1:size(R,2));
        zq = Qn'*zvec_wQ;
        x3 = Un\zq;
        
        %Calculation of normalised residuals
        r_Q = zvec_Q-H.Q*x3;
        Ohm = diag(z_Q(:,5))-H.Q*Ginv3*H.Q';
        r_Q_N = abs(r_Q)./sqrt(abs(diag(Ohm)));
        [r_Q_N_max, j] = max(r_Q_N);
    end
    
    if contV
        contV = false;
        nz_V = size(z_V,1);
        H.V =  zeros(nz_V, n_V);
        
        for k = 1:nz_V
            if z_V(k,3) == 1 %if conventional voltage measurement
                H.V(k,z_V(k,1)) = 1;
            else %if CB pseudo-measurement
                H.V(k,z_V(k,1)) = 1;
                H.V(k,z_V(k,2)) = -1;
            end
        end
        
        %Cholesky factorisation solution method
        zvec_V = z_V(:,4);
        Ginv1 = ( H.V' * ((diag(z_V(:,5))) \ H.V )) \ eye(n_V);
        x1 = Ginv1*H.V'*(diag(z_V(:,5))\zvec_V);
        
        %Calculation of normalised residuals
        r_V = zvec_V-H.V*x1;
        Ohm = diag(z_V(:,5))-H.V*Ginv1*H.V';
        r_V_N = abs(r_V)./sqrt(abs(diag(Ohm)));
        [r_V_N_max, i_V] = max(r_V_N);
    end
    
    %Bad data detection
    P_BDdetected = r_P_N_max > 3; %alternative, the chi square criterion: r_P'*(diag(z_P(:,5))\r_P) > chi2inv(0.99,length(r_P)-n_P);
    Q_BDdetected = r_Q_N_max > 3; %alternative, the chi square criterion: r_Q'*(diag(z_Q(:,5))\r_Q) > chi2inv(0.99,length(r_Q)-n_Q);
    
    if (P_BDdetected && ~Q_BDdetected && r_P_N_max > 3) || (P_BDdetected && r_P_N_max > r_Q_N_max && r_P_N_max > 3) %error in active power
        if  z_P(i_P,3) == 12 %incorrect open breaker
            z_Q(z_Q(:,1) == z_P(i_P,1) & z_Q(:,2) == z_P(i_P,2) & z_Q(:,3) == 13, :) = []; %removes closed breaker constraint on Q also
            if display
                fprintf('Removed open breaker topology error in substation %4g:', s)
                disp(z_P(i_P,1:2));
            end
            
            %Enforcing the opposite CB pseudo-measurement:
            z_V = [z_V; z_P(i_P, 1:2), 11 ,z_P(i_P, 4:5)];
            nz_V = nz_V +1;
            z_P(i_P, :) = [];
            z_Q(1,:) = [];
            contP = true;
            contQ = true;
            contV = true;
        else %bad analogue data
            if display
                fprintf('Removed bad data in substation %4g:', s)
                disp(z_P(i_P,:));
            end
            z_P(i_P,:)=[];
            contP = true;
        end
        
    elseif (Q_BDdetected && ~P_BDdetected && r_Q_N_max > 3) || (Q_BDdetected && r_Q_N_max > r_P_N_max && r_Q_N_max > 3) %error in reactive power
        if z_Q(j,3) == 13 %incorrect open breaker
            z_P(z_P(:,1) == z_Q(j,1) & z_P(:,2) == z_Q(j,2) & z_P(:,3) == 13, :) = [];  %removes closed breaker constraint on P also
            if display
                fprintf('Removed open breaker topology error in substation %4g:', s)
                disp(z_Q(j,1:2));
            end
            
            %Enforcing the opposite CB pseudo-measurement:
            z_V = [z_V; z_Q(j, 1:2), 11 ,z_Q(j, 4:5)];
            nz_V = nz_V +1;
            z_Q(j, :) = [];
            contP = true;
            contQ = true;
            contV = true;
        else  %bad analogue data
            if display
                fprintf('Removed bad data in substation %4g:', s)
                disp(z_Q(j,:));
            end
            z_Q(j,:)=[];
            contQ = true;
        end
        
    elseif r_V_N_max > 3
         if z_V(i_V,3) == 11 %incorrect closed breaker
            if display
                fprintf('Removed closed breaker topology error in substation %4g:', s)
                disp(z_V(i_ind,1:2));
            end
            
            %Enforcing the opposite CB pseudo-measurement:
            z_P = [z_P; z_P(i_V, 1:2), 12 ,z_P(i_V, 4:5)];
            z_Q = [z_Q; z_Q(i_V, 1:2), 13 ,z_Q(i_V, 4:5)];
            nz_P = nz_P +1;
            nz_Q = nz_Q +1;
            z_V(i_V, :) = [];
            contP = true;
            contQ = true;
            contV = true;
         else %bad analogue data
            if display
                fprintf('Removed bad data in substation %4g:', s)
                disp(z_V(i_V,:));
            end
            z_V(i_V,:)=[];
            contV = true;
        end
    end
end


%% Bulding the pre-processed measurements of the central SE
Vbus = mean(x1);
Vbus_sigma = mean(diag(Ginv1));

%Building H.Pinj and H.Qinj matrices:
    H.Pinj = zeros(n_nodes,1);  %injected power / incoming power flow on every node
    for k = 1:n_nodes
        H.Pinj(k,  (k) == localbranches(:,1)) = 1;
        H.Pinj(k,  (k) == localbranches(:,2)) = -1;
    end

%Calculating out to central SE
Pinj = H.Pinj*x2;
Pinj_sigma = H.Pinj*Ginv2*H.Pinj';

%% Building final measurement matrix from current substation:
index1 = find(extlinedata(:, 1) == s);
index2 = find(extlinedata(:, 2) == s);
zcentral = zeros(3+2*(length(index1)+length(index2)), 5);


%Active power flows:
for i = 1:length(index1)
    zcentral(2+2*i,:) = [s,  extlinedata(index1(i), 2),  4, -Pinj(extlinedata(index1(i), 3)), Pinj_sigma(extlinedata(index1(i), 3), extlinedata(index1(i), 3))];
end
for i = 1:length(index2)
    zcentral(2+2*i+2*length(index1),:) = [s, extlinedata(index2(i), 1), 4, -Pinj(extlinedata(index2(i), 4)), Pinj_sigma(extlinedata(index2(i), 4), extlinedata(index2(i), 4))];
end





H.Qinj = zeros(n_nodes,1);
for k = 1:n_nodes
    H.Qinj(k,  (k) == localbranches(:,1)) = 1;
    H.Qinj(k,  (k) == localbranches(:,2)) = -1;
end

Qinj = H.Qinj*x3;
Qinj_sigma = H.Qinj*Ginv3*H.Qinj';
%include phase angles?

%Reactive power flows:
for i = 1:length(index1)
    zcentral(3+2*i,:) = [ s, extlinedata(index1(i), 2), 5, -Qinj(extlinedata(index1(i), 3)), Qinj_sigma(extlinedata(index1(i), 3), extlinedata(index1(i), 3))];
end
for i = 1:length(index2)
    zcentral(3+2*i+2*length(index1),:) = [s, extlinedata(index2(i), 1),  5, -Qinj(extlinedata(index2(i), 4)), Qinj_sigma(extlinedata(index2(i), 4), extlinedata(index2(i), 4))];
end

%Voltage
zcentral(1,:) = [s, 0, 1, Vbus, Vbus_sigma];
%zcentral(1,:) = [s, 0, 1, mean(zlocal(zlocal(:,3)==1, 4)), mean(zlocal(zlocal(:,3)==1, 5))/length(zlocal(zlocal(:,3)==1, 5))];

%Power injections:
ind = nodetype((nodetype(:,2)==1 | nodetype(:,2)==2 | nodetype(:,2)==3),1);
zcentral(2,:) = [s, 0, 2, sum(Pinj(ind)), max([10^-12, sum(sum(Pinj_sigma(ind,ind)))])];

ind = nodetype((nodetype(:,2)==1 | nodetype(:,2)==2 | nodetype(:,2)==3),1);
zcentral(3,:) = [s, 0, 3, sum(Qinj(ind)), max([10^-12, sum(sum(Qinj_sigma(ind,ind)))])];

end